"""This task runs gersemi on CMake files."""

import subprocess
import sys

from wpiformat.config import Config
from wpiformat.task import BatchTask


class CMakeFormat(BatchTask):
    @staticmethod
    def should_process_file(config_file: Config, filename: str) -> bool:
        return filename.endswith("CMakeLists.txt") or filename.endswith(".cmake")

    @staticmethod
    def run_batch(config_file: Config, filenames: list[str]) -> bool:
        try:
            args = [sys.executable, "-m", "gersemi", "-i", "--no-color", "-q"]
            subprocess.run(args + filenames)
        except FileNotFoundError:
            print("error: gersemi not found in PATH. Is it installed?", file=sys.stderr)
            return False
        return True



from wpiformat.usingdeclaration import UsingDeclaration

from .test_tasktest import *


def test_usingdeclaration():
    # Before class block
    run_and_check_stdout(
        UsingDeclaration(),
        "./Test.h",
        """using std::chrono;
class Test {
}
""",
        "./Test.h: 1: 'using std::chrono;' in global namespace\n",
        False,
    )

    # Inside enum block
    run_and_check_stdout(
        UsingDeclaration(),
        "./Test.h",
        """enum Test {
  using std::chrono;
}
""",
        "",
        True,
    )

    # After { block
    run_and_check_stdout(
        UsingDeclaration(),
        "./Test.h",
        """{
}
using std::chrono;
""",
        "./Test.h: 3: 'using std::chrono;' in global namespace\n",
        False,
    )

    # Before class block with NOLINT
    run_and_check_stdout(
        UsingDeclaration(),
        "./Test.h",
        """using std::chrono;  // NOLINT
class Test {
}
""",
        "",
        True,
    )

    # "using" in comment without trailing semicolon
    run_and_check_stdout(
        UsingDeclaration(),
        "./Test.h",
        """// using
void func() {
  using A = int;
  using B = int;
}
""",
        "",
        True,
    )



"""This task ensures that the file has zero EOF newlines if it's empty or one EOF newline."""

from wpiformat.config import Config
from wpiformat.task import PipelineTask


class EofNewline(PipelineTask):
    def run_pipeline(
        self, config_file: Config, filename: str, lines: str
    ) -> tuple[str, bool]:
        lines = lines.rstrip()
        if lines:
            return lines + super().get_linesep(lines), True
        else:
            return lines, True



"""This task runs clang-format on the file."""

import subprocess
import sys

import clang_format

from wpiformat.config import Config
from wpiformat.task import PipelineTask


class ClangFormat(PipelineTask):
    def __init__(self):
        """Constructor for ClangFormat task."""
        super().__init__()

        self.exec_name = clang_format.get_executable("clang-format")

    @staticmethod
    def should_process_file(config_file: Config, filename: str) -> bool:
        return config_file.is_c_file(filename) or config_file.is_cpp_file(filename)

    def run_pipeline(
        self, config_file: Config, filename: str, lines: str
    ) -> tuple[str, bool]:
        try:
            p = subprocess.Popen(
                [self.exec_name, "-style=file", f"-assume-filename={filename}", "-"],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                encoding="utf-8",
            )
            stdout, stderr = p.communicate(input=lines)

            if p.returncode != 0:
                print(
                    f"error: {self.exec_name} returned non-zero exit status {p.returncode}",
                    file=sys.stderr,
                )
                print(stderr, file=sys.stderr)
                return lines, False
        except FileNotFoundError:
            print(
                f"error: {self.exec_name} not found in PATH. Is it installed?",
                file=sys.stderr,
            )
            return lines, False

        return stdout, True



from wpiformat.whitespace import Whitespace

from .test_tasktest import *


def test_whitespace():
    file_appendix = """#pragma once

#include <iostream>

int main() {
  std::cout << "Hello World!";
}
"""

    # Empty file
    run_and_check_file(Whitespace(), "./Test.h", "", "", True)

    # No trailing whitespace
    run_and_check_file(Whitespace(), "./Test.h", file_appendix, file_appendix, True)

    # Two spaces trailing
    run_and_check_file(
        Whitespace(),
        "./Test.h",
        "#pragma once\n"
        + "\n"
        + "#include <iostream>\n"
        + "\n"
        + "int main() {  \n"
        + '  std::cout << "Hello World!";  \n'
        + "}\n",
        file_appendix,
        True,
    )

    # Two tabs trailing
    run_and_check_file(
        Whitespace(),
        "./Test.h",
        """#pragma once

#include <iostream>

int main() {\t\t
  std::cout << "Hello World!";\t\t
}
""",
        file_appendix,
        True,
    )



"""This task ensures the trailing comment on the closing brace of extern and
namespace declarations matches that of the declaration name.
"""

import re

from wpiformat.config import Config
from wpiformat.task import PipelineTask


class BraceComment(PipelineTask):
    @staticmethod
    def should_process_file(config_file: Config, filename: str) -> bool:
        return config_file.is_c_file(filename) or config_file.is_cpp_file(filename)

    def run_pipeline(
        self, config_file: Config, filename: str, lines: str
    ) -> tuple[str, bool]:
        linesep = super().get_linesep(lines)
        output = ""

        brace_prefix = r"(?P<prefix>(extern|namespace)\s+[\w\"]*)"
        brace_postfix = r"[ \t]*/(/|\*)[^\r\n]*"

        brace_regex = re.compile(
            r"/\*|\*/|//|\\\\|\\\"|\"|\\'|'|"
            + linesep
            + r"|"
            + r"("
            + brace_prefix
            + r"\s*)?{|"  # "{" with optional prefix
            r"}(" + brace_postfix + r")?"
        )  # "}" with optional comment postfix

        name_stack = []
        brace_count = 0
        extract_location = 0
        in_multicomment = False
        in_singlecomment = False
        in_string = False
        in_char = False
        for match in brace_regex.finditer(lines):
            token = match.group()

            if token == "/*":
                if not in_singlecomment and not in_string and not in_char:
                    in_multicomment = True
            elif token == "*/":
                if not in_singlecomment and not in_string and not in_char:
                    in_multicomment = False
            elif token == "//":
                if not in_multicomment and not in_string and not in_char:
                    in_singlecomment = True
            elif in_singlecomment and linesep in token:
                # Ignore token if it's in a singleline comment. Only check it
                # for newlines to end the comment.
                in_singlecomment = False
            elif in_multicomment or in_singlecomment:
                # Tokens processed after this branch are ignored if they are in
                # comments
                continue
            elif token == '\\"':
                continue
            elif token == '"':
                if not in_char:
                    in_string = not in_string
            elif token == "\\'":
                continue
            elif token == "'":
                if not in_string:
                    in_char = not in_char
            elif in_string or in_char:
                # Tokens processed after this branch are ignored if they are in
                # double or single quotes
                continue
            elif match.group("prefix"):
                brace_count += 1
                name_stack.append((brace_count, match.group("prefix").rstrip()))
            elif "{" in token:
                brace_count += 1
            elif token.startswith("}"):
                output += lines[extract_location : match.start()]
                if (
                    len(name_stack) > 0
                    and name_stack[len(name_stack) - 1][0] == brace_count
                ):
                    # If there's a line continuation, use a multiline comment
                    # instead
                    end_of_line = lines.find(linesep, match.start())
                    rest_of_line = lines[match.start() : end_of_line]
                    if rest_of_line.endswith("\\"):
                        output += f"}}  /* {name_stack.pop()[1]} */ "
                        extract_location = end_of_line - 1
                    else:
                        output += f"}}  // {name_stack.pop()[1]}"
                        extract_location = match.end()
                else:
                    output += lines[match.start() : match.end()]
                    extract_location = match.end()
                brace_count -= 1

        # If input has unprocessed lines, write them to output
        if extract_location < len(lines):
            output += lines[extract_location:]

        return output, True



"""This class is for writing unit tests for tasks."""

import io
import os
from contextlib import redirect_stdout

from wpiformat.config import Config
from wpiformat.task import Task


def run_and_check_file(
    task: Task,
    filename: str,
    input_contents: str,
    expected_output_contents: str,
    expected_success: bool,
):
    """
    Runs the task on the input file contents, then compares the resulting output
    against the expected output file contents.

    Keyword Arguments:
    task -- Task instance to test
    filename -- filename
    input_contents -- input file contents
    expected_output_contents -- expected output file contents
    expected_success -- whether run is expected to succeed
    """
    input_contents = input_contents.replace("\n", os.linesep)
    expected_output_contents = expected_output_contents.replace("\n", os.linesep)

    config_file = Config(os.path.abspath(os.getcwd()), ".styleguide")

    if task.should_process_file(config_file, filename):
        output, success = task.run_pipeline(config_file, filename, input_contents)
    else:
        output = input_contents
        success = True

    assert output == expected_output_contents
    assert success == expected_success


def run_and_check_stdout(
    task: Task,
    filename: str,
    input_contents: str,
    expected_output_contents: str,
    expected_success: bool,
):
    """
    Runs the task on the input file contents, then compares the resulting
    output against the expected stdout contents.

    Keyword Arguments:
    task -- Task instance to test
    filename -- filename
    input_contents -- input file contents
    expected_output_contents -- expected output file contents
    expected_success -- whether run is expected to succeed
    """
    config_file = Config(os.path.abspath(os.getcwd()), ".styleguide")

    if task.should_process_file(config_file, filename):
        with redirect_stdout(io.StringIO()) as f:
            _, success = task.run_pipeline(config_file, filename, input_contents)
        output = f.getvalue()
    else:
        output = input_contents
        success = True

    assert output == expected_output_contents
    assert success == expected_success



"""This task removes trailing whitespace from the file."""

from wpiformat.config import Config
from wpiformat.task import PipelineTask


class Whitespace(PipelineTask):
    def run_pipeline(
        self, config_file: Config, filename: str, lines: str
    ) -> tuple[str, bool]:
        linesep = super().get_linesep(lines)

        return "".join(line.rstrip() + linesep for line in lines.splitlines()), True



from wpiformat.gtestname import GTestName

from .test_tasktest import *


def test_gtestname():
    # Ignore TEST macro that isn't for GoogleTest
    contents = "AXIS_TEST(Joystick, X)\n"
    run_and_check_file(GTestName(), "./Test.cpp", contents, contents, True)

    for test_type in ["TEST", "TEST_F", "TEST_P"]:
        # Test suite replacements
        run_and_check_file(
            GTestName(),
            "./Test.cpp",
            f"{test_type}(Foo, Bar) {{}}\n",
            f"{test_type}(FooTest, Bar) {{}}\n",
            True,
        )
        run_and_check_file(
            GTestName(),
            "./Test.cpp",
            f"{test_type}(FooTests, Bar) {{}}\n",
            f"{test_type}(FooTest, Bar) {{}}\n",
            True,
        )

        # Test case replacements
        run_and_check_file(
            GTestName(),
            "./Test.cpp",
            f"{test_type}(FooTest, BarTest) {{}}\n",
            f"{test_type}(FooTest, Bar) {{}}\n",
            True,
        )

        # Test case only named "Test" or "Tests" should generate an error
        contents = f"{test_type}(FooTest, Test) {{}}\n"
        run_and_check_file(GTestName(), "./Test.cpp", contents, contents, False)
        contents = f"{test_type}(FooTest, Tests) {{}}\n"
        run_and_check_file(GTestName(), "./Test.cpp", contents, contents, False)

        # Idempotency
        contents = f"{test_type}(FooTest, Bar) {{}}\n"
        run_and_check_file(GTestName(), "./Test.cpp", contents, contents, True)

        # Duplicate test names to make sure file contents aren't dropped
        run_and_check_file(
            GTestName(),
            "./Test.cpp",
            f"""{test_type}(FooTest, BarTest) {{}}

{test_type}(FooTest, BarTest) {{}}
""",
            f"""{test_type}(FooTest, Bar) {{}}

{test_type}(FooTest, Bar) {{}}
""",
            True,
        )

        # Test case replacement with args on next line
        run_and_check_file(
            GTestName(),
            "./Test.cpp",
            f"""{test_type}(
    FooTest, BarTest) {{}}
""",
            f"""{test_type}(
    FooTest, Bar) {{}}
""",
            True,
        )

    # INSTANTIATE_TEST_SUITE_P() test suite replacements
    run_and_check_file(
        GTestName(),
        "./Test.cpp",
        "INSTANTIATE_TEST_SUITE_P(Foo, BarTest,\n",
        "INSTANTIATE_TEST_SUITE_P(FooTests, BarTest,\n",
        True,
    )
    run_and_check_file(
        GTestName(),
        "./Test.cpp",
        "INSTANTIATE_TEST_SUITE_P(FooTest, BarTest,\n",
        "INSTANTIATE_TEST_SUITE_P(FooTests, BarTest,\n",
        True,
    )

    # INSTANTIATE_TEST_SUITE_P() test case replacements
    run_and_check_file(
        GTestName(),
        "./Test.cpp",
        "INSTANTIATE_TEST_SUITE_P(FooTests, Bar,\n",
        "INSTANTIATE_TEST_SUITE_P(FooTests, BarTest,\n",
        True,
    )
    run_and_check_file(
        GTestName(),
        "./Test.cpp",
        "INSTANTIATE_TEST_SUITE_P(FooTests, BarTests,\n",
        "INSTANTIATE_TEST_SUITE_P(FooTests, BarTest,\n",
        True,
    )

    # INSTANTIATE_TEST_SUITE_P() idempotency
    contents = "INSTANTIATE_TEST_SUITE_P(FooTests, BarTest,\n"
    run_and_check_file(GTestName(), "./Test.cpp", contents, contents, True)

    # INSTANTIATE_TEST_SUITE_P() test case replacement with args on next line
    run_and_check_file(
        GTestName(),
        "./Test.cpp",
        """INSTANTIATE_TEST_SUITE_P(
    FooTests, BarTests,
""",
        """INSTANTIATE_TEST_SUITE_P(
    FooTests, BarTest,
""",
        True,
    )



from wpiformat.javaclass import JavaClass

from .test_tasktest import *


def test_javaclass():
    # No line separators at beginning of class
    run_and_check_file(
        JavaClass(),
        "./Test.java",
        "public class ExampleCommand extends Command {}\n",
        "public class ExampleCommand extends Command {}\n",
        True,
    )

    # One line separator at beginning of class
    contents = """public class ExampleCommand extends Command {
  public ExampleCommand() {}
}
"""
    run_and_check_file(JavaClass(), "./Test.java", contents, contents, True)

    # Two line separators at beginning of class
    run_and_check_file(
        JavaClass(),
        "./Test.java",
        """public class ExampleCommand extends Command {

  public ExampleCommand() {}
}
""",
        """public class ExampleCommand extends Command {
  public ExampleCommand() {}
}
""",
        True,
    )

    # Three line separators at beginning of class
    run_and_check_file(
        JavaClass(),
        "./Test.java",
        """public class ExampleCommand extends Command {


  public ExampleCommand() {}
}
""",
        """public class ExampleCommand extends Command {
  public ExampleCommand() {}
}
""",
        True,
    )

    # class keyword in preceding comment to ensure regex matching it doesn't
    # continue past end of comment
    run_and_check_file(
        JavaClass(),
        "./Test.java",
        """import edu.wpi.first.networktables.NetworkTableEntry;
import edu.wpi.first.wpilibj.Sendable;

/**
 * A helper class for Shuffleboard containers to handle common child operations.
 */
final class ContainerHelper {

  private final ShuffleboardContainer m_container;
""",
        """import edu.wpi.first.networktables.NetworkTableEntry;
import edu.wpi.first.wpilibj.Sendable;

/**
 * A helper class for Shuffleboard containers to handle common child operations.
 */
final class ContainerHelper {
  private final ShuffleboardContainer m_container;
""",
        True,
    )



"""Task base classes for wpiformat."""

import os
import subprocess
from abc import ABCMeta, abstractmethod

from wpiformat.config import Config


class Task(metaclass=ABCMeta):
    @staticmethod
    def get_linesep(lines: str) -> str:
        """Returns string containing autodetected line separator for file.

        Keyword arguments:
        lines -- file contents
        """
        # Find potential line separator
        pos = lines.find("\n")

        # If a newline character was found and the character preceding it is a
        # carriage return, assume CRLF line endings. LF line endings are assumed
        # for empty files.
        if pos > 0 and lines[pos - 1] == "\r":
            return "\r\n"
        else:
            return "\n"

    @staticmethod
    def get_repo_root() -> str:
        """Returns the Git repository root as an absolute path.

        An empty string is returned if no repository root was found.
        """
        return os.path.normpath(
            subprocess.check_output(
                ["git", "rev-parse", "--show-toplevel"], encoding="ascii"
            ).rstrip()
        )

    @staticmethod
    def should_process_file(config_file: Config, filename: str) -> bool:
        """Returns true if file should be processed by this task.

        Keyword arguments:
        config_file -- Config object
        name -- filename

        Process any file by default.
        """
        return True


class PipelineTask(Task):
    @abstractmethod
    def run_pipeline(
        self, config_file: Config, filename: str, lines: str
    ) -> tuple[str, bool]:
        """Performs task on file with given lines.

        This function is for processing the file in a pipeline of tasks.

        Keyword arguments:
        config_file -- Config object
        name -- filename
        lines -- file contents

        Returns tuple containing processed lines and whether task succeeded in
        processing the file.
        """
        return "", True


class BatchTask(Task):
    @staticmethod
    @abstractmethod
    def run_batch(config_file: Config, filenames: list[str]) -> bool:
        """Performs task on list of files.

        This function is for processing multiple files in one task to reduce
        overhead.

        Keyword arguments:
        config_file -- Config object
        filenames -- list of filenames

        Returns True if task succeeded in processing the files.
        """
        return True


class StandaloneTask(Task):
    @abstractmethod
    def run_standalone(self, config_file: Config, filename: str) -> bool:
        """Performs task on a file.

        This function is for processing the file on its own.

        Keyword arguments:
        config_file -- Config object
        filename -- filename

        Returns True if task succeeded in processing the file.
        """
        return True



from wpiformat.cidentlist import CIdentList

from .test_tasktest import *


def test_cidentlist():
    # Main.cpp: signature for C++ function
    contents = """int main() {
  return 0;
}
"""
    run_and_check_file(CIdentList(), "./Main.cpp", contents, contents, True)

    # Main.cpp: signature for C function in extern "C" block
    run_and_check_file(
        CIdentList(),
        "./Main.cpp",
        """extern "C" {
int main() {
  return 0;
}
}
""",
        """extern "C" {
int main(void) {
  return 0;
}
}
""",
        True,
    )

    # Main.cpp: signature for C function marked extern "C"
    run_and_check_file(
        CIdentList(),
        "./Main.cpp",
        """extern "C" int main() {
  return 0;
}
""",
        """extern "C" int main(void) {
  return 0;
}
""",
        True,
    )

    # Main.cpp: extern "C++" function in extern "C" block
    contents = """extern "C" {
extern "C++" int main() {
  return 0;
}
}
"""
    run_and_check_file(CIdentList(), "./Main.cpp", contents, contents, True)

    # Main.c: signature for C function
    run_and_check_file(
        CIdentList(),
        "./Main.c",
        """int main() {
  return 0;
}
""",
        """int main(void) {
  return 0;
}
""",
        True,
    )

    # Main.c: signature for C++ function in extern "C++" block
    contents = """extern "C++" {
int main() {
  return 0;
}
}
"""
    run_and_check_file(CIdentList(), "./Main.c", contents, contents, True)

    # Main.c: signature for C++ function marked extern "C++"
    contents = """extern "C++" int main() {
  return 0;
}
"""
    run_and_check_file(CIdentList(), "./Main.c", contents, contents, True)

    # Main.c: extern "C" function nested in extern "C++" block
    run_and_check_file(
        CIdentList(),
        "./Main.c",
        """extern "C++" {
extern "C" int main() {
  return 0;
}
}
""",
        """extern "C++" {
extern "C" int main(void) {
  return 0;
}
}
""",
        True,
    )

    # Don't match function calls
    run_and_check_file(
        CIdentList(),
        "./Main.c",
        """int main() {
  foo();
  return 0;
}
""",
        """int main(void) {
  foo();
  return 0;
}
""",
        True,
    )

    # Don't match function calls with return (return is a keyword not a return
    # type)
    run_and_check_file(
        CIdentList(),
        "./Main.c",
        """int main() {
  return foo();
}
""",
        """int main(void) {
  return foo();
}
""",
        True,
    )

    # Match function prototypes
    run_and_check_file(
        CIdentList(),
        "./Main.c",
        """int main();

int main() {
  foo();
  return 0;
}
""",
        """int main(void);

int main(void) {
  foo();
  return 0;
}
""",
        True,
    )

    # Make sure leaving extern block resets extern language type of parent block
    run_and_check_file(
        CIdentList(),
        "./Main.c",
        """extern "C++" {
extern "C" int main() {
  return 0;
}
int func() {
  return 0;
}
}
""",
        """extern "C++" {
extern "C" int main(void) {
  return 0;
}
int func() {
  return 0;
}
}
""",
        True,
    )

    # Don't match lambda function that takes no arguments
    contents = """extern "C" {

HAL_Bool HAL_Initialize(int32_t timeout, int32_t mode) {
  std::atexit([]() {
    // Unregister our new data condition variable.
  });
}

}  // extern "C"
"""
    run_and_check_file(CIdentList(), "./Main.cpp", contents, contents, True)

    # Don't match a function call within a #ifdef
    contents = """ES_Event Elevator_Service_Run(ES_Event event) {
#ifdef USE_TATTLETALE
    ES_Tail(); // trace call stack end
#endif
}
"""
    run_and_check_file(CIdentList(), "./Main.c", contents, contents, True)

    run_and_check_file(
        CIdentList(),
        "./Timer.hpp",
        """extern "C" void Timer1IntHandler();

class Timer {
public:
    void Set(uint32_t newTime);

    void Start();

    void Stop();

    uint16_t GetID() const;

    static uint32_t GetTime();

private:
    friend void Timer1IntHandler();
};
""",
        """extern "C" void Timer1IntHandler(void);

class Timer {
public:
    void Set(uint32_t newTime);

    void Start();

    void Stop();

    uint16_t GetID() const;

    static uint32_t GetTime();

private:
    friend void Timer1IntHandler();
};
""",
        True,
    )

    # Ensure comments with } in them don't mess up brace stack
    contents = """void func() {
  // closing }
}
"""
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure nested comments don't mess up brace stack
    contents = "// { /* */ }\n"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)
    contents = "{ // /* */ }\n"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, False)
    contents = "{ /* // */ }\n"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)
    contents = "{ /* */ // }\n"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, False)
    contents = "{ // /* // */ }\n"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, False)

    # Ensure popping too many braces doesn't crash
    run_and_check_file(CIdentList(), "./Test.cpp", "}\n", "}\n", False)

    # Ensure comments inside quoted string don't mess up brace stack
    contents = """void func() {
  // "//"
  if (!query.startswith("//")) {
    return;
  }
}
"""
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure braces in double quotes don't mess up brace stack
    contents = "void func() { std::cout << '{'; }"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure braces in single quotes don't mess up brace stack
    contents = 'void func() { std::cout << "{"; }'
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure single quote within double quotes doesn't mess up brace stack
    contents = 'void func() { std::cout << "\'"; }'
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure double quote within single quotes doesn't mess up brace stack
    contents = "void func() { std::cout << '\"'; }"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure escaped double quote doesn't mess up brace stack
    contents = 'void func() { std::cout << "\\""; }'
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure escaped single quote doesn't mess up brace stack
    contents = "void func() { std::cout << '\\''; }"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure escaped backslash isn't considered as an escaped single quote
    contents = "void func() { std::cout << '\\\\'; }"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure extern "C" match containing a linesep within a singleline comment
    # still ends the comment
    contents = """extern "C" {}  // extern "C"
namespace {
}  // namespace
"""
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure extern "C" with brace on next line gets matched
    run_and_check_file(
        CIdentList(),
        "./Test.cpp",
        """extern "C"
{
  void func() {}
}  // extern "C"
""",
        """extern "C"
{
  void func(void) {}
}  // extern "C"
""",
        True,
    )

    # Test logic for deduplicating braces within #ifdef
    contents = """void func() {
#ifdef _WIN32
  if (errno == WSAEWOULDBLOCK) {
#else
  if (errno == EWOULDBLOCK) {
#endif
  }
}
"""
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure extern "C" function with pointer return type gets matched
    run_and_check_file(
        CIdentList(),
        "./Test.cpp",
        'extern "C" void* func() {}\n',
        'extern "C" void* func(void) {}\n',
        True,
    )

    # Ensure single quotes in numeric literals are ignored
    contents = "void func() { int x = 1'000; }"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure single quotes in hexadecimal literals are ignored
    contents = "void func() { int x = 0xffff'ffff; }"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure single quotes after numeric literals are not ignored
    contents = "void func() { std::cout << 1 << '0'; }"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)

    # Ensure single quotes after hexadecimal characters are not ignored
    contents = "void func() { std::cout << 1 << 'a'; }"
    run_and_check_file(CIdentList(), "./Test.cpp", contents, contents, True)



import os

from wpiformat.includeguard import IncludeGuard
from wpiformat.task import Task

from .test_tasktest import *


def test_includeguard():
    repo_root = os.path.basename(Task.get_repo_root()).upper()

    # Fix incorrect include guard
    run_and_check_file(
        IncludeGuard(),
        "./Test.h",
        """#ifndef WRONG_H
#define WRONG_C

#endif
""",
        f"""#ifndef {repo_root}_TEST_H_
#define {repo_root}_TEST_H_

#endif  // {repo_root}_TEST_H_
""",
        True,
    )

    # Ensure nested preprocessor statements are handled properly for incorrect
    # include guard
    run_and_check_file(
        IncludeGuard(),
        "./Test.h",
        """#ifndef WRONG_H
#define WRONG_C

#if SOMETHING
// do something
#endif
#endif
""",
        f"""#ifndef {repo_root}_TEST_H_
#define {repo_root}_TEST_H_

#if SOMETHING
// do something
#endif
#endif  // {repo_root}_TEST_H_
""",
        True,
    )

    # Don't touch correct include guard
    contents = f"""#ifndef {repo_root}_TEST_H_
#define {repo_root}_TEST_H_

#endif  // {repo_root}_TEST_H_
"""
    run_and_check_file(IncludeGuard(), "./Test.h", contents, contents, True)

    # Fail on missing include guard
    run_and_check_file(
        IncludeGuard(), "./Test.h", "// Empty file\n", "// Empty file\n", False
    )

    # Verify pragma once counts as include guard
    run_and_check_file(
        IncludeGuard(), "./Test.h", "#pragma once\n", "#pragma once\n", True
    )

    # Ensure include guard roots are processed correctly
    run_and_check_file(
        IncludeGuard(),
        "./Test.h",
        f"""#ifndef {repo_root}_WPIFORMAT_TEST_H_
#define {repo_root}_WPIFORMAT_TEST_H_

#endif  // {repo_root}_WPIFORMAT_TEST_H_
""",
        f"""#ifndef {repo_root}_TEST_H_
#define {repo_root}_TEST_H_

#endif  // {repo_root}_TEST_H_
""",
        True,
    )

    # Ensure leading underscores are removed (this occurs if the user doesn't
    # include a trailing "/" in the include guard root)
    run_and_check_file(
        IncludeGuard(),
        "./Test/Test.h",
        f"""#ifndef {repo_root}_WPIFORMAT_TEST_TEST_H_
#define {repo_root}_WPIFORMAT_TEST_TEST_H_

#endif  // {repo_root}_WPIFORMAT_TEST_TEST_H_
""",
        f"""#ifndef {repo_root}_TEST_H_
#define {repo_root}_TEST_H_

#endif  // {repo_root}_TEST_H_
""",
        True,
    )



from wpiformat.usingnamespacestd import UsingNamespaceStd

from .test_tasktest import *


def test_usingnamespacestd():
    warning_str = 'avoid "using namespace std;" in production software. While it is used in introductory C++, it pollutes the global namespace with standard library symbols. Be more specific and use "using std::thing;" instead.\n'

    # Hello World
    run_and_check_stdout(
        UsingNamespaceStd(),
        "./Main.cpp",
        """using namespace std;

int main() {
  cout << "Hello World!"
}
""",
        "warning: ./Main.cpp: 1: " + warning_str,
        True,
    )

    # Inside braces and not first line
    run_and_check_stdout(
        UsingNamespaceStd(),
        "./Main.cpp",
        """int main() {
  using namespace std;
  cout << "Hello World!"
}
""",
        "warning: ./Main.cpp: 2: " + warning_str,
        True,
    )

    # std::chrono
    run_and_check_stdout(
        UsingNamespaceStd(),
        "./Main.cpp",
        """#include <thread>

int main() {
  using namespace std::chrono;

  std::this_thread::sleep_for(10ms);
}
""",
        "warning: ./Main.cpp: 4: " + warning_str,
        True,
    )

    # Ignore std::literals
    run_and_check_stdout(
        UsingNamespaceStd(), "./Main.cpp", "using namespace std::literals;\n", "", True
    )

    # Ignore std::chrono_literals
    run_and_check_stdout(
        UsingNamespaceStd(),
        "./Main.cpp",
        "using namespace std::chrono_literals;\n",
        "",
        True,
    )

    # Ignore std::string_view_literals
    run_and_check_stdout(
        UsingNamespaceStd(),
        "./Main.cpp",
        "using namespace std::string_view_literals;\n",
        "",
        True,
    )

    # Ignore std::placeholders
    run_and_check_stdout(
        UsingNamespaceStd(),
        "./Main.cpp",
        "using namespace std::placeholders;\n",
        "",
        True,
    )



"""This task runs black on files with Python extension."""

import subprocess
import sys

from wpiformat.config import Config
from wpiformat.task import BatchTask


class PyFormat(BatchTask):
    @staticmethod
    def should_process_file(config_file: Config, filename: str) -> bool:
        return filename.endswith(".py")

    @staticmethod
    def run_batch(config_file: Config, filenames: list[str]) -> bool:
        try:
            args = [
                sys.executable,
                "-m",
                "autoflake",
                "--remove-all-unused-imports",
                "--remove-unused-variables",
                "--ignore-init-module-imports",
                "--quiet",
                "-i",
            ]
            subprocess.run(args + filenames)
        except FileNotFoundError:
            print(
                "error: autoflake not found in PATH. Is it installed?", file=sys.stderr
            )
            return False

        try:
            args = [sys.executable, "-m", "black", "-q"]
            subprocess.run(args + filenames)
        except FileNotFoundError:
            print("error: black not found in PATH. Is it installed?", file=sys.stderr)
            return False

        try:
            args = [sys.executable, "-m", "isort", "--profile", "black", "-q"]
            subprocess.run(args + filenames)
        except FileNotFoundError:
            print("error: isort not found in PATH. Is it installed?", file=sys.stderr)
            return False

        return True



import os

from wpiformat.eofnewline import EofNewline

from .test_tasktest import *


def test_eofnewline():
    file_appendix = """#pragma once

#include <iostream>

int main() {
  std::cout << "Hello World!";
}"""
    test_output = f"{file_appendix}\n"

    # Empty file
    run_and_check_file(EofNewline(), "./Test.h", "", "", True)

    # No newline
    run_and_check_file(EofNewline(), "./Test.h", file_appendix, test_output, True)

    # One newline
    run_and_check_file(EofNewline(), "./Test.h", test_output, test_output, True)

    # Two newlines
    run_and_check_file(EofNewline(), "./Test.h", f"{test_output}\n", test_output, True)

    # .bat file with no "./" prefix
    if os.linesep == "\r\n":
        run_and_check_file(EofNewline(), "test.bat", file_appendix, test_output, True)
    else:
        run_and_check_file(
            EofNewline(),
            "test.bat",
            file_appendix.replace("\n", "\r\n"),
            test_output.replace("\n", "\r\n"),
            True,
        )



from wpiformat.jni import Jni

from .test_tasktest import *


def test_jni():
    # Input args go to next line even if they fit on same line
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT void JNICALL
Java_TestJNI_testFunc(JNIEnv* env, jclass) {
""",
        """/*
 * Class:     TestJNI
 * Method:    testFunc
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_TestJNI_testFunc
  (JNIEnv* env, jclass)
{
""",
        True,
    )

    # Input aligned to "(" and args past end of line
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT void JNICALL
Java_edu_wpi_cscore_CameraServerJNI_setCameraExposureHoldCurrent(JNIEnv* env,
                                                                 jclass,
                                                                 jint source) {
""",
        """/*
 * Class:     edu_wpi_cscore_CameraServerJNI
 * Method:    setCameraExposureHoldCurrent
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_edu_wpi_cscore_CameraServerJNI_setCameraExposureHoldCurrent
  (JNIEnv* env, jclass, jint source)
{
""",
        True,
    )

    # Args in input on line after "(" and args length > 80 characters
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT void JNICALL Java_edu_wpi_cscore_CameraServerJNI_putSourceFrame(
    JNIEnv *env, jclass, jint source, jlong imageNativeObj) {
""",
        """/*
 * Class:     edu_wpi_cscore_CameraServerJNI
 * Method:    putSourceFrame
 * Signature: (IJ)V
 */
JNIEXPORT void JNICALL
Java_edu_wpi_cscore_CameraServerJNI_putSourceFrame
  (JNIEnv *env, jclass, jint source, jlong imageNativeObj)
{
""",
        True,
    )

    # Args > 80 characters long
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT jint JNICALL Java_edu_wpi_cscore_CameraServerJNI_createSourceProperty(
    JNIEnv *env, jclass, jint source, jstring name, jint kind, jint minimum,
    jint maximum, jint step, jint defaultValue, jint value) {
""",
        """/*
 * Class:     edu_wpi_cscore_CameraServerJNI
 * Method:    createSourceProperty
 * Signature: (ILjava/lang/String;IIIIII)I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_cscore_CameraServerJNI_createSourceProperty
  (JNIEnv *env, jclass, jint source, jstring name, jint kind, jint minimum,
   jint maximum, jint step, jint defaultValue, jint value)
{
""",
        True,
    )

    # Ensure fixes clang-format output aligned with "("
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance(JNIEnv*,
                                                                 jclass) {
""",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv*, jclass)
{
""",
        True,
    )

    # Idempotence for same code
    contents = """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv*, jclass)
{
"""
    run_and_check_file(Jni(), "./TestJNI.cpp", contents, contents, True)

    # Idempotence for same code with named jclass variable
    contents = """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv*, jclass class)
{
"""
    run_and_check_file(Jni(), "./TestJNI.cpp", contents, contents, True)

    # Check signature that breaks verbose regexes
    run_and_check_file(
        Jni(),
        "./NetworkTablesJNI.cpp",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getEntry
 * Signature: (ILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getEntry(JNIEnv* env, jclass,
                                                           jint inst,
                                                           jstring key) {
""",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getEntry
 * Signature: (ILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getEntry
  (JNIEnv* env, jclass, jint inst, jstring key)
{
""",
        True,
    )

    # Function with array type as argument
    run_and_check_file(
        Jni(),
        "./NetworkTablesJNI.cpp",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getEntries
 * Signature: (ILjava/lang/String;I)[I
 */
JNIEXPORT jintArray JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getEntries(JNIEnv* env,
                                                             jclass, jint inst,
                                                             jstring prefix,
                                                             jint types) {
""",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getEntries
 * Signature: (ILjava/lang/String;I)[I
 */
JNIEXPORT jintArray JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getEntries
  (JNIEnv* env, jclass, jint inst, jstring prefix, jint types)
{
""",
        True,
    )

    # Ensure functions with overloads are handled correctly
    contents = """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    setRaw
 * Signature: (IJ[BZ)Z
 */
JNIEXPORT jboolean JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_setRaw__IJ_3BZ
  (JNIEnv* env, jclass, jint entry, jlong time, jbyteArray value,
   jboolean force)
{
"""
    run_and_check_file(Jni(), "./NetworkTablesJNI.cpp", contents, contents, True)

    # Ensure text before JNIEXPORT and after args and ")" is handled correctly
    # as well as two JNI functions in a row
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """/**
 *
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getDefaultInstance
  (JNIEnv *, jclass)
{
  return nt::GetDefaultInstance();
}

JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv *, jclass)
{
  return nt::CreateInstance();
}
""",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getDefaultInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getDefaultInstance
  (JNIEnv *, jclass)
{
  return nt::GetDefaultInstance();
}

/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv *, jclass)
{
  return nt::CreateInstance();
}
""",
        True,
    )

    # Handle function declarations properly
    contents = """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getDefaultInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getDefaultInstance
  (JNIEnv *, jclass);

/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv *, jclass)
{
  return nt::CreateInstance();
}
"""
    run_and_check_file(Jni(), "./TestJNI.cpp", contents, contents, True)

    # Handle functions whose arguments don't have variable names properly
    run_and_check_file(
        Jni(),
        "./DigitalGlitchFilterJNI.cpp",
        """/*
 * Class:     edu_wpi_first_wpilibj_hal_DigitalGlitchFilterJNI
 * Method:    cleanFilter
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_edu_wpi_first_wpilibj_hal_DigitalGlitchFilterJNI_cleanFilter
  (JNIEnv *, jclass, jint)
{
  HAL_CleanFilter(handle);
}
""",
        """/*
 * Class:     edu_wpi_first_wpilibj_hal_DigitalGlitchFilterJNI
 * Method:    cleanFilter
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_edu_wpi_first_wpilibj_hal_DigitalGlitchFilterJNI_cleanFilter
  (JNIEnv *, jclass, jint)
{
  HAL_CleanFilter(handle);
}
""",
        True,
    )

    # Input args go to next line even if they fit on same line
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT void JNICALL
Java_TestJNI_testFunc(JNIEnv* env, jobject) {
""",
        """/*
 * Class:     TestJNI
 * Method:    testFunc
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_TestJNI_testFunc
  (JNIEnv* env, jobject)
{
""",
        True,
    )

    # Input aligned to "(" and args past end of line
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT void JNICALL
Java_edu_wpi_cscore_CameraServerJNI_setCameraExposureHoldCurrent(JNIEnv* env,
                                                                 jobject,
                                                                 jint source) {
""",
        """/*
 * Class:     edu_wpi_cscore_CameraServerJNI
 * Method:    setCameraExposureHoldCurrent
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_edu_wpi_cscore_CameraServerJNI_setCameraExposureHoldCurrent
  (JNIEnv* env, jobject, jint source)
{
""",
        True,
    )

    # Args in input on line after "(" and args length > 80 characters
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT void JNICALL Java_edu_wpi_cscore_CameraServerJNI_putSourceFrame(
    JNIEnv *env, jobject, jint source, jlong imageNativeObj) {
""",
        """/*
 * Class:     edu_wpi_cscore_CameraServerJNI
 * Method:    putSourceFrame
 * Signature: (IJ)V
 */
JNIEXPORT void JNICALL
Java_edu_wpi_cscore_CameraServerJNI_putSourceFrame
  (JNIEnv *env, jobject, jint source, jlong imageNativeObj)
{
""",
        True,
    )

    # Args > 80 characters long
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT jint JNICALL Java_edu_wpi_cscore_CameraServerJNI_createSourceProperty(
    JNIEnv *env, jobject, jint source, jstring name, jint kind, jint minimum,
    jint maximum, jint step, jint defaultValue, jint value) {
""",
        """/*
 * Class:     edu_wpi_cscore_CameraServerJNI
 * Method:    createSourceProperty
 * Signature: (ILjava/lang/String;IIIIII)I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_cscore_CameraServerJNI_createSourceProperty
  (JNIEnv *env, jobject, jint source, jstring name, jint kind, jint minimum,
   jint maximum, jint step, jint defaultValue, jint value)
{
""",
        True,
    )

    # Ensure fixes clang-format output aligned with "("
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance(JNIEnv*,
                                                                 jobject) {
""",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv*, jobject)
{
""",
        True,
    )

    # Idempotence for same code
    contents = """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv*, jobject)
{
"""
    run_and_check_file(Jni(), "./TestJNI.cpp", contents, contents, True)

    # Idempotence for same code with named jobject variable
    contents = """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv*, jobject class)
{
"""
    run_and_check_file(Jni(), "./TestJNI.cpp", contents, contents, True)

    # Check signature that breaks verbose regexes
    run_and_check_file(
        Jni(),
        "./NetworkTablesJNI.cpp",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getEntry
 * Signature: (ILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getEntry(JNIEnv* env, jobject,
                                                           jint inst,
                                                           jstring key) {
""",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getEntry
 * Signature: (ILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getEntry
  (JNIEnv* env, jobject, jint inst, jstring key)
{
""",
        True,
    )

    # Function with array type as argument
    run_and_check_file(
        Jni(),
        "./NetworkTablesJNI.cpp",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getEntries
 * Signature: (ILjava/lang/String;I)[I
 */
JNIEXPORT jintArray JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getEntries(JNIEnv* env,
                                                             jobject, jint inst,
                                                             jstring prefix,
                                                             jint types) {
""",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getEntries
 * Signature: (ILjava/lang/String;I)[I
 */
JNIEXPORT jintArray JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getEntries
  (JNIEnv* env, jobject, jint inst, jstring prefix, jint types)
{
""",
        True,
    )

    # Ensure functions with overloads are handled correctly
    contents = """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    setRaw
 * Signature: (IJ[BZ)Z
 */
JNIEXPORT jboolean JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_setRaw__IJ_3BZ
  (JNIEnv* env, jobject, jint entry, jlong time, jbyteArray value,
   jboolean force)
{
"""
    run_and_check_file(Jni(), "./NetworkTablesJNI.cpp", contents, contents, True)

    # Ensure text before JNIEXPORT and after args and ")" is handled correctly
    # as well as two JNI functions in a row
    run_and_check_file(
        Jni(),
        "./TestJNI.cpp",
        """/**
 *
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getDefaultInstance
  (JNIEnv *, jobject)
{
  return nt::GetDefaultInstance();
}

JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv *, jobject)
{
  return nt::CreateInstance();
}
""",
        """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getDefaultInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getDefaultInstance
  (JNIEnv *, jobject)
{
  return nt::GetDefaultInstance();
}

/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv *, jobject)
{
  return nt::CreateInstance();
}
""",
        True,
    )

    # Handle function declarations properly
    contents = """/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    getDefaultInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_getDefaultInstance
  (JNIEnv *, jobject);

/*
 * Class:     edu_wpi_first_networktables_NetworkTablesJNI
 * Method:    createInstance
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_edu_wpi_first_networktables_NetworkTablesJNI_createInstance
  (JNIEnv *, jobject)
{
  return nt::CreateInstance();
}
"""
    run_and_check_file(Jni(), "./TestJNI.cpp", contents, contents, True)

    # Handle functions whose arguments don't have variable names properly
    run_and_check_file(
        Jni(),
        "./DigitalGlitchFilterJNI.cpp",
        """/*
 * Class:     edu_wpi_first_wpilibj_hal_DigitalGlitchFilterJNI
 * Method:    cleanFilter
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_edu_wpi_first_wpilibj_hal_DigitalGlitchFilterJNI_cleanFilter
  (JNIEnv *, jobject, jint)
{
  HAL_CleanFilter(handle);
}
""",
        """/*
 * Class:     edu_wpi_first_wpilibj_hal_DigitalGlitchFilterJNI
 * Method:    cleanFilter
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_edu_wpi_first_wpilibj_hal_DigitalGlitchFilterJNI_cleanFilter
  (JNIEnv *, jobject, jint)
{
  HAL_CleanFilter(handle);
}
""",
        True,
    )



"""This task runs cpplint.py on a C++ source file."""

import os
import sys
from contextlib import redirect_stdout

import cpplint

from wpiformat.config import Config
from wpiformat.task import BatchTask


class Lint(BatchTask):
    @staticmethod
    def should_process_file(config_file: Config, filename: str) -> bool:
        return config_file.is_cpp_file(filename)

    @staticmethod
    def run_batch(config_file: Config, filenames: list[str]) -> bool:
        # Prepare arguments to cpplint.py
        saved_argv = sys.argv

        exclusion_filters = [
            "build/c++11",
            "build/c++17",
            "build/header_guard",
            "build/include_order",
            "build/include_subdir",
            "build/namespaces",
            "legal/copyright",
            "readability/braces",
            "readability/check",
            "readability/todo",
            "runtime/references",
            "runtime/string",
            "whitespace/braces",
            "whitespace/comma",
            "whitespace/comments",
            "whitespace/end_of_line",
            "whitespace/ending_newline",
            "whitespace/indent",
            "whitespace/indent_namespace",
            "whitespace/line_length",
            "whitespace/newline",
            "whitespace/operators",
            "whitespace/parens",
            "whitespace/semicolon",
            "whitespace/tab",
        ]

        # Prepare header file extensions
        header_exts = []
        for pattern in config_file.group("cHeaderFileInclude") + config_file.group(
            "cppHeaderFileInclude"
        ):
            basename = os.path.basename(pattern)
            header_exts.append(basename[basename.rfind(".") + 1 :].rstrip("$"))

        args = ["cpplint.py", "--filter=-" + ",-".join(exclusion_filters)]
        if header_exts:
            args.append("--headers=" + ",".join(header_exts))
        args.append("--quiet")
        sys.argv = args + filenames

        # Run cpplint.py
        try:
            with open(os.devnull, "w") as devnull, redirect_stdout(devnull):
                cpplint.main()
        except SystemExit as e:
            # Restore original arguments
            sys.argv = saved_argv

            # Report success if error code is 0 (False)
            return e.code == False
        return False



"""This task makes include guards follow the style stipulated by the Google
style guide.
"""

import os
import re
from enum import Enum

from wpiformat.config import Config
from wpiformat.task import PipelineTask


class State(Enum):
    FINDING_IFNDEF = 1
    FINDING_ENDIF = 2
    DONE = 3


class IncludeGuard(PipelineTask):
    @staticmethod
    def should_process_file(config_file: Config, filename: str):
        return config_file.is_header_file(filename)

    def run_pipeline(
        self, config_file: Config, filename: str, lines: str
    ) -> tuple[str, bool]:
        linesep = super().get_linesep(lines)
        lines_list = lines.split(linesep)
        output_list = lines_list

        state = State.FINDING_IFNDEF
        ifndef_regex = re.compile(r"#ifndef \w+", re.ASCII)
        define_regex = re.compile(r"#define \w+", re.ASCII)

        if_preproc_count = 0
        for i in range(len(lines_list)):
            if state == State.FINDING_IFNDEF:
                if lines_list[i].lstrip().startswith("#ifndef ") and lines_list[
                    i + 1
                ].lstrip().startswith("#define "):
                    state = State.FINDING_ENDIF

                    guard = self.make_include_guard(config_file, filename)
                    output_list[i] = ifndef_regex.sub("#ifndef " + guard, lines_list[i])
                    output_list[i + 1] = define_regex.sub(
                        "#define " + guard, lines_list[i + 1]
                    )
                    if_preproc_count += 1
                elif lines_list[i].lstrip().startswith("#pragma once"):
                    state = State.DONE
            elif state == State.FINDING_ENDIF:
                if "#if" in lines_list[i]:
                    if_preproc_count += 1
                elif "#endif" in lines_list[i]:
                    if_preproc_count -= 1

                if if_preproc_count == 0:
                    state = State.DONE
                    output_list[i] = "#endif  // " + guard
                else:
                    output_list[i] = lines_list[i]
            else:
                output_list[i] = lines_list[i]

        # If include guard not found
        if state == State.FINDING_IFNDEF:
            print(f"error: {filename}: doesn't contain include guard or '#pragma once'")
            return lines, False

        output = linesep.join(output_list).rstrip() + linesep
        return output, True

    def make_include_guard(self, config_file: Config, filename: str) -> str:
        """Returns properly formatted include guard based on repository root and
        filename.

        Keyword arguments:
        config_file -- Config object
        filename -- filename
        """
        repo_root_name_override = config_file.group("repoRootNameOverride")

        repo_root = super().get_repo_root()
        guard_root = os.path.relpath(filename, repo_root)
        if not repo_root_name_override:
            guard_path = os.path.basename(repo_root) + os.sep
        else:
            guard_path = repo_root_name_override[0] + os.sep
        include_roots = config_file.group("includeGuardRoots")

        if include_roots:
            prefix = ""
            for include_root in include_roots:
                if guard_root.startswith(include_root) and len(include_root) > len(
                    prefix
                ):
                    prefix = include_root
            guard_path += guard_root[len(prefix) :]
            return (re.sub(r"[^a-zA-Z0-9]", "_", guard_path).upper() + "_").lstrip("_")

        # No include guard roots matched, so append full name
        guard_path += guard_root
        return re.sub(r"[^a-zA-Z0-9]", "_", guard_path).upper() + "_"



import os
import subprocess
import tempfile
from datetime import date
from pathlib import Path

from wpiformat.config import Config
from wpiformat.licenseupdate import LicenseUpdate

from .test_tasktest import *


class OpenTemporaryDirectory:
    def __init__(self):
        self.prev_dir = os.getcwd()

    def __enter__(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        os.chdir(self.temp_dir.name)
        return self.temp_dir

    def __exit__(self, type, value, traceback):
        os.chdir(self.prev_dir)


def test_licenseupdate():
    year = str(date.today().year)

    file_appendix = """#pragma once

#include <iostream>

int main() {
  std::cout << "Hello World!";
}"""

    # pragma once at top of file
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        file_appendix,
        f"""/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */

"""
        + file_appendix,
        True,
    )

    # pragma once at top of file preceded by newline
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        "\n" + file_appendix,
        f"""/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */

"""
        + file_appendix,
        True,
    )

    # File containing up-to-date license preceded by newline
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        f"""

/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */

"""
        + file_appendix,
        f"""/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */

"""
        + file_appendix,
        True,
    )

    # File containing up-to-date range license
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        f"""/*                                Company Name                                */
// Copyright (c) 2011-{year} Company Name. All Rights Reserved.

"""
        + file_appendix,
        f"""/*                                Company Name                                */
/* Copyright (c) 2011-{year} Company Name. All Rights Reserved.                 */

"""
        + file_appendix,
        True,
    )

    # File containing up-to-date license with one year
    contents = (
        f"""/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */

"""
        + file_appendix
    )
    run_and_check_file(LicenseUpdate(), "./Test.h", contents, contents, True)

    # File with three newlines between license and include guard
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        f"""/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */


"""
        + file_appendix,
        f"""/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */

"""
        + file_appendix,
        True,
    )

    # File with only one newline between license and include guard
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        f"""/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */
"""
        + file_appendix,
        f"""/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */

"""
        + file_appendix,
        True,
    )

    # File with multiline comment spanning multiple lines of license header
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        """/*
 * Autogenerated file! Do not manually edit this file. This version is regenerated
 * any time the publish task is run, or when this file is deleted.
 */

const char* WPILibVersion = "";""",
        f"""/*                                Company Name                                */
/* Copyright (c) {year} Company Name. All Rights Reserved.                      */

/*
 * Autogenerated file! Do not manually edit this file. This version is regenerated
 * any time the publish task is run, or when this file is deleted.
 */

const char* WPILibVersion = "";""",
        True,
    )

    # File containing license year range in different postion than template
    # (If the year isn't extracted, the range will be replaced with one year and
    # the test will fail.)
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        f"""/*                                Company Name                                */
/* Copyright (c) Company Name 2011-{year}. All Rights Reserved.                 */

"""
        + file_appendix,
        f"""/*                                Company Name                                */
/* Copyright (c) 2011-{year} Company Name. All Rights Reserved.                 */

"""
        + file_appendix,
        True,
    )

    # Ensure "/*" after "*/" on same line is detected
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        """/*----------------------------------------------------------------------------*/
/* Copyright (c) 2011 FIRST. All Rights Reserved.                             */
/* Open Source Software - may be modified and shared by FRC teams. The code   */
/* must be accompanied by the FIRST BSD license file in the root directory of */
/* the project.                                                               */
/*----------------------------------------------------------------------------*//*

blah

*/
""",
        f"""/*                                Company Name                                */
/* Copyright (c) 2011-{year} Company Name. All Rights Reserved.                 */

""",
        True,
    )

    # File excluded from license update isn't modified
    contents = f"""/* Copyright (c) Company Name 2011-{year}. */

"""
    run_and_check_file(LicenseUpdate(), "./Excluded.h", contents, contents, True)

    # Ensure license regex matches
    run_and_check_file(
        LicenseUpdate(),
        "./Test.h",
        """/* Company Name */
/* Copyright (c) 1992-2015 Company Name. All Rights Reserved. */

"""
        + file_appendix,
        f"""/*                                Company Name                                */
/* Copyright (c) 1992-{year} Company Name. All Rights Reserved.                 */

"""
        + file_appendix,
        True,
    )

    # Ensure excluded files won't be processed
    config_file = Config(os.path.abspath(os.getcwd()), ".styleguide")
    assert not LicenseUpdate().should_process_file(config_file, "./Excluded.h")

    # Create git repo to test license years for commits
    with OpenTemporaryDirectory():
        subprocess.run(["git", "init", "-q"])

        # Add base files
        with open(".styleguide-license", "w") as file:
            file.write("// Copyright (c) {year}")
        with open(".styleguide", "w") as file:
            file.write("cppSrcFileInclude {\n" + r"\.cpp$")
        subprocess.run(["git", "add", ".styleguide-license"])
        subprocess.run(["git", "add", ".styleguide"])
        subprocess.run(["git", "commit", "-q", "-m", '"Initial commit"'])

        # Add file with commit date of last year and range through this year
        with open("last-year.cpp", "w") as file:
            file.write(f"// Copyright (c) 2017-{year}")
        subprocess.run(["git", "add", "last-year.cpp"])
        subprocess.run(["git", "commit", "-q", "-m", '"Last year"'])
        last_iso_year = f"{int(year) - 1}-01-01T00:00:00"
        subprocess.Popen(
            ["git", "commit", "-q", "--amend", "--no-edit", f"--date={last_iso_year}"],
            env={**os.environ, "GIT_COMMITTER_DATE": last_iso_year},
        ).wait()

        # Add file with commit date of this year and range through this year
        with open("this-year.cpp", "w") as file:
            file.write(f"// Copyright (c) 2017-{year}")
        subprocess.run(["git", "add", "this-year.cpp"])
        subprocess.run(["git", "commit", "-q", "-m", '"This year"'])

        # Add file with commit date of next year and range through this year
        with open("next-year.cpp", "w") as file:
            file.write(f"// Copyright (c) 2017-{year}")
        subprocess.run(["git", "add", "next-year.cpp"])
        subprocess.run(["git", "commit", "-q", "-m", '"Next year"'])
        next_iso_year = f"{int(year) + 1}-01-01T00:00:00"
        subprocess.Popen(
            ["git", "commit", "-q", "--amend", "--no-edit", f"--date={next_iso_year}"],
            env={**os.environ, "GIT_COMMITTER_DATE": next_iso_year},
        ).wait()

        # Create uncommitted file with no year
        Path("no-year.cpp").touch()

        # Run wpiformat on last-year.cpp
        with open("last-year.cpp", "r") as input:
            lines = input.read()
        output, success = LicenseUpdate().run_pipeline(
            config_file, "last-year.cpp", lines
        )
        assert output == f"// Copyright (c) 2017-{int(year) - 1}\n\n"

        # Run wpiformat on last-year.cpp with uncommitted changes. It should
        # update to next year instead of keeping previous year
        with open("last-year.cpp", "a") as input:
            input.write("change\n")
        output, success = LicenseUpdate().run_pipeline(
            config_file, "last-year.cpp", lines + "change\n"
        )
        assert (
            output
            == f"""// Copyright (c) 2017-{year}

change
"""
        )

        # Erase changes made to last-year.cpp in previous test
        with open("last-year.cpp", "w") as input:
            input.write(lines)

        # Run wpiformat on this-year.cpp
        with open("last-year.cpp", "r") as input:
            lines = input.read()
        output, success = LicenseUpdate().run_pipeline(
            config_file, "this-year.cpp", lines
        )
        assert (
            output
            == f"""// Copyright (c) 2017-{year}

"""
        )

        # Run wpiformat on next-year.cpp
        with open("next-year.cpp", "r") as input:
            lines = input.read()
        output, success = LicenseUpdate().run_pipeline(
            config_file, "next-year.cpp", lines
        )
        assert (
            output
            == f"""// Copyright (c) 2017-{int(year) + 1}

"""
        )

        # Run wpiformat on no-year.cpp
        # Should have current calendar year
        with open("no-year.cpp", "r") as input:
            lines = input.read()
        output, success = LicenseUpdate().run_pipeline(
            config_file, "no-year.cpp", lines
        )
        assert (
            output
            == f"""// Copyright (c) {year}

"""
        )



from wpiformat.bracecomment import BraceComment

from .test_tasktest import *


def test_bracecomment():
    # Empty anonymous namespace
    run_and_check_file(
        BraceComment(),
        "./Test.h",
        """namespace {
}// comment
""",
        """namespace {
}  // namespace
""",
        True,
    )

    # Anonymous namespace containing comment
    run_and_check_file(
        BraceComment(),
        "./Test.h",
        """namespace {
  // comment
}// comment
""",
        """namespace {
  // comment
}  // namespace
""",
        True,
    )

    # namespace
    run_and_check_file(
        BraceComment(),
        "./Test.h",
        """namespace hal {
  // comment
}// comment
""",
        """namespace hal {
  // comment
}  // namespace hal
""",
        True,
    )

    # namespace with leftover input
    run_and_check_file(
        BraceComment(),
        "./Test.h",
        """// comment before namespace
namespace hal {
  // comment
}// comment
// comment after namespace
""",
        """// comment before namespace
namespace hal {
  // comment
}  // namespace hal
// comment after namespace
""",
        True,
    )

    # Braces within namespace
    run_and_check_file(
        BraceComment(),
        "./Test.h",
        """namespace {

struct AnalogGyro {
  HAL_AnalogInputHandle handle;
  double voltsPerDegreePerSecond;
  double offset;
  int32_t center;
}

}
""",
        """namespace {

struct AnalogGyro {
  HAL_AnalogInputHandle handle;
  double voltsPerDegreePerSecond;
  double offset;
  int32_t center;
}

}  // namespace
""",
        True,
    )

    # extern "C"
    run_and_check_file(
        BraceComment(),
        "./Test.h",
        """extern "C" {
    // nothing
}// comment
""",
        """extern "C" {
    // nothing
}  // extern "C"
""",
        True,
    )

    # Nested brackets should be handled properly
    contents = """extern "C" {
void func() {
  if (1) {
  } else if (1) {
  } else {
  }
}
}  // extern "C"
"""
    run_and_check_file(BraceComment(), "./Test.cpp", contents, contents, True)

    # Nested brackets on same line
    run_and_check_file(
        BraceComment(),
        "./Test.cpp",
        """namespace wpi {
{{}}
}  // namespace java
""",
        """namespace wpi {
{{}}
}  // namespace wpi
""",
        True,
    )

    # Handle single-line statements correctly
    run_and_check_file(
        BraceComment(),
        "./Test.cpp",
        "namespace hal { Type typeName; }\n",
        "namespace hal { Type typeName; }  // namespace hal\n",
        True,
    )

    # Two incorrect comments
    run_and_check_file(
        BraceComment(),
        "./Test.h",
        """namespace {
    // nothing
}// comment
namespace Name {
    // nothing
}
""",
        """namespace {
    // nothing
}  // namespace
namespace Name {
    // nothing
}  // namespace Name
""",
        True,
    )

    # Don't touch correct comment
    contents = """namespace {
    // nothing
}  // namespace
"""
    run_and_check_file(BraceComment(), "./Test.h", contents, contents, True)

    # Handle braces in comments properly
    run_and_check_file(
        BraceComment(),
        "./Path.h",
        """#ifndef ALLWPILIB_WPI_PATH_H_
#define ALLWPILIB_WPI_PATH_H_

namespace wpi {
namespace sys {
namespace path {

/// @{

}  // end namespace path
}  // namespace sys
}  // namespace wpi

#endif  // ALLWPILIB_WPI_PATH_H_
""",
        """#ifndef ALLWPILIB_WPI_PATH_H_
#define ALLWPILIB_WPI_PATH_H_

namespace wpi {
namespace sys {
namespace path {

/// @{

}  // namespace path
}  // namespace sys
}  // namespace wpi

#endif  // ALLWPILIB_WPI_PATH_H_
""",
        True,
    )

    # Comment in macro
    run_and_check_file(
        BraceComment(),
        "./Test.cpp",
        """#define TEST(namespaceName, name, ...) \\
  namespace namespaceName { \\
  using name = __VA_ARGS__; \\
  } \\
  using name = namespaceName::name;
""",
        """#define TEST(namespaceName, name, ...) \\
  namespace namespaceName { \\
  using name = __VA_ARGS__; \\
  }  /* namespace namespaceName */ \\
  using name = namespaceName::name;
""",
        True,
    )
    run_and_check_file(
        BraceComment(),
        "./Test.cpp",
        """#define TEST(namespaceName, name, ...) \\
  namespace namespaceName { \\
  using name = __VA_ARGS__; \\
  } /* namespaceName */\\
  using name = namespaceName::name;
""",
        """#define TEST(namespaceName, name, ...) \\
  namespace namespaceName { \\
  using name = __VA_ARGS__; \\
  }  /* namespace namespaceName */ \\
  using name = namespaceName::name;
""",
        True,
    )
